
/******************************************************************
		kaboom.c

   This file contains all of the routines for fiber flexure
   geometry and collision avoidance.


******************************************************************/

#include <stdio.h>
/* #include <sys/time.h> */

#include "miscMath.h"
#include "plate.h"
#include "motor.h"
#include "kaboom.h"
#include "occult.h"

#ifdef ARCHIVE
#include "archive.h"
#endif

#ifdef LOG
#include "log.h"
#endif

extern char *malloc();

/************************  GLOBALS  *****************************/

struct fiber_dbase FiBase[MAXFIBER];	/*  master database */

struct snapShot FiberInTransit;  /*  button that is being moved */

int softAngleFlag = 0;

/********************* local variables **************************/

static double radiusList[(MAXFIBER / 2)];

/* fractional points along a tube to choose endpoints of line
   segments */

static double segConstants[] = {
   
    SEGPT0,
    SEGPT1,
    SEGPT2,
    SEGPT3,
    SEGPT4,
};


/********************* local prototypes **************************/

static int    buttonExtension( fibDB *fiber, struct flex_geometry *geometry );
static int    couldInteract( fibDB *fiber1, fibDB *fiber2 );
static int    buttonsConverge( fibDB *fib1, fibDB *fib2, fibDB **lowFib, fibDB **hiFib, double *thetaSep);
static int    buttonHitsTube( fibDB *targetFib, fibDB *lowFib, fibDB *hiFib, int convergeStatus, double sepAngle);
static double fiberButtonDistance( struct flex_geometry *buttonGeometry, struct flex_geometry *fiberGeometry, double angle );
static void   constructLine( struct flex_geometry *geometry, double dN, line *theLine );
static int    tubesCross( fibDB *targetFib, fibDB *lowFib, fibDB *hiFib, int   convergeStatus, double sepAngle);
static int    miscOtherMagic( fibDB *lowFib, fibDB *hiFib, int convergeStatus, double sepAngle);
static int    buttonSeparation( fibDB *fiber1, fibDB *fiber2 );
static void   tubeSlope( struct flex_geometry *flex, double dN, double *localSlope, double *cartesianSlope) ;
static void   tubeLocation( struct flex_geometry *flex, double dN, point *thePoint );
static void   tubeAngle( struct flex_geometry *flex, double dN, double *phiN );
static void   MagicNumbers( struct flex_geometry *results );

/********************* end local prototypes **************************/


/*
		pendingDelta

   Compute the size of the impending move for the specified button.

*/

double pendingDelta( fibDB *pFib )

{
   return( distance( &(pFib->CurGeometry->button),
		     &(pFib->TrialGeometry->button) ) );
}



/*

		DestOK

   This is the major collision avoidance routine.
   It first checks to be sure that the destination is within
   the range of the fiber, then checks for collisions with other
   fibers.  Fibers crossing is also detected.  Any of these
   conditions will cause the routine to return false.
   The functions exits as soon as an error is detected.
*/


int DestOK( fibDB fibers[], fibDB *fib )

{
   fibDB *targetFib;

   int statusCode;
   int i;


   /* see if the fiber is occulted */


   if( isOcculted( fib ) ){
      occultMessage( fib );
      goto kaboom;
   }


   /* see of the range is OK */

   statusCode = coordStatus( fib, NO_RESTRICT );

   switch( statusCode ){

      case OVER_EXTEND :
	 
#ifdef LOG
         log_fprintf(stderr, "Move would over flex/extend fiber %d\n", fib->fiber_id);
#else
         fprintf(stderr, "Move would over flex/extend fiber %d\n", fib->fiber_id);
#endif
	 goto kaboom;
	 break;

      case ON_STEP     :

#ifdef LOG
	 log_fputs( "Requested move is in the forbidden zone\n", stderr );
#else
	 fputs( "Requested move is in the forbidden zone\n", stderr );
#endif

	 goto kaboom;
	 break;
   }

   /* clear the pending occultation list in case it was set previously and not recorded */

   occultClearPending( fib );

   /* loop through all other fibers and check for collisions */

   for( i=0; i< MAXFIBER; i++){
      if( i != fib->fiber_id ){   	/* don't do current fiber! */

	 targetFib = &fibers[i];

	 statusCode = pairHits( fib, targetFib );

	 switch( statusCode ){


	    case BUTTON_BUTTON :
	       
#ifdef LOG
	       log_fprintf( stderr, "button %d hits button %d!\n",
			 fib->fiber_id, targetFib->fiber_id );
#else
	       fprintf( stderr, "button %d hits button %d!\n",
			 fib->fiber_id, targetFib->fiber_id );
#endif
	       goto kaboom;
	       break;


	    case BUTTON_TUBE  :

#ifdef LOG
      	       log_fprintf( stderr, "%d and %d interact button/tube!\n",
     			 fib->fiber_id, targetFib->fiber_id );
#else
      	       fprintf( stderr, "%d and %d interact button/tube!\n",
     			 fib->fiber_id, targetFib->fiber_id );
#endif
	       goto kaboom;
	       break;


	    case TUBE_TUBE  :

#ifdef LOG
	       log_fprintf( stderr, "Tube %d crosses tube %d. \n",
			 fib->fiber_id, targetFib->fiber_id );
#else
	       fprintf( stderr, "Tube %d crosses tube %d. \n",
			 fib->fiber_id, targetFib->fiber_id );
#endif
	       goto kaboom;
	       break;

	 }
      }
   }

   /* if we make it through the loop and reach here, there
	is no collision and we return */

   return( TRUE );  /* remember: returns destination "ok" */

   kaboom: return( FALSE );
}



/*
		thisIsOK

   This function is called by "thisis" to be sure that the user is not
   mis-identifying a button. 
*/

int thisIsOK( fibDB fibers[], fibDB *fib )

{
   struct fiber_dbase *targetFib;

   int safe = TRUE;
   int i;


   /* loop through all other fibers and check for collisions */

   for( i=0; i< MAXFIBER; i++){
      if( i != fib->fiber_id ){   	/* don't do current fiber! */

	 targetFib = &fibers[i];

	 if( buttonSeparation( fib, targetFib ) != 0 ){
#ifdef LOG
	    log_fprintf( stderr,
	       "WARNING: button %d is *very* nearby!\n", targetFib->fiber_id );
#else
	    fprintf( stderr,
	       "WARNING: button %d is *very* nearby!\n", targetFib->fiber_id );
#endif
	 }
	 if( overButton( targetFib, &(fib->CurGeometry->button) ) ){

#ifdef LOG
	    log_fprintf( stderr,
	    "ERROR: database thinks this location is over button %d!!!\n", targetFib->fiber_id );
#else
	    fprintf( stderr,
	    "ERROR: database thinks this location is over button %d!!!\n", targetFib->fiber_id );
#endif

	    safe = FALSE;
	 }
      }
   }

   return( safe );  
}



/*
		coordStatus

   See if the desired location is acceptable for the 
   specified fiber.  Return a code for the type of 
   error if it is not.
*/

int coordStatus( fibDB *fib, int restrictFlag )

{
   int whichStep;

   /* check for over-extension  */

   if( buttonExtension( fib, fib->CurGeometry ) ){
      return( OVER_EXTEND );
   }

   if( PlateZ( fib, &(fib->CurGeometry->button), &whichStep, HARD_BOUNDRIES) == PLATE_BAD ){
      return( ON_STEP );
   }

   if( restrictFlag && ( whichStep == PARKORD ) ){
      return( OVER_EXTEND );
   }
   
   return( 0 );  /* no error */
}



/*
		pairHits

   See if the pair of fibers specified interact in any way.
   Return a code describing the type of collision if they do.
   Break out as soon as collision is detected.
*/


int pairHits( fibDB *fib, fibDB *targetFib )

{
   int collisonCode = 0;

   struct fiber_dbase *lowFib;
   struct fiber_dbase *hiFib;

   int convergeStatus;
   double sepAngle;

   /*   Check for button/button collision */

   if( buttonSeparation( fib, targetFib ) ){
      collisonCode = BUTTON_BUTTON;
      goto end;
   }

   /* see if an interaction is possible by first
      running some fast, simple tests */

   if( !couldInteract(fib, targetFib) )
      goto end;   /* can't possibly collide */

   
   /* Determine if tubes are diverging or converging.
      Figure out which tube is CW wrt the angle between them.
      If they diverge, then we are done.
      If they converge, things get hairy. */

   convergeStatus = buttonsConverge( fib, targetFib,
				   &lowFib, &hiFib, &sepAngle );

   if( miscOtherMagic( lowFib, hiFib,
		convergeStatus, sepAngle )  == NOHELP ){

      /*   Check for button/tube collision */

      if( buttonHitsTube( fib, lowFib, hiFib,
		   convergeStatus, sepAngle ) ){
	 collisonCode = BUTTON_TUBE;
	 goto end;
      }

      /*   Check for tube/tube collision */

      if( tubesCross( fib, lowFib, hiFib,
		   convergeStatus, sepAngle ) ){
	 collisonCode = TUBE_TUBE;
	 goto end;
      }
   }

   /* if we fall through it  tells us that
      the two buttons/fibers won't collide. */

   end: return( collisonCode );
	
}



/*
		buttonExtension
   Determine if the target destination is within the allowable
   range for the specified button. Return zero if the button
   is within range.
*/

int buttonExtension( fibDB *fiber, struct flex_geometry *geometry )

{
   if( ((long)geometry->d4 > fiber->MaxExtend) ||
       (((long)geometry->d2 > PLATE ) && ((long)geometry->d2 < STOW ) ) ){

       return( TRUE );
   }
  

   if( softAngleFlag == TRUE ){
      if( fabs(geometry->psi) > ( fiber->MaxAngle + ANGLE_SOFTNESS) )
	 return(TRUE );
   }
   else{
      if( fabs(geometry->psi) > fiber->MaxAngle )
	 return(TRUE );
   }

   return( FALSE );
}
   


/*
      PlateZ

   Find the distance to the plate at the current gripper location.
   If a step falls within a button radius of the target location,
   return zero to indicate an indeterminate result.

   The flag is used to tell the function whether the software intends
   to pick up or drop a button at the specified coordinate.  
   Grabbing buttons requires no buffer zone so that it can't be
   possible to have a button in an unreachable location. and thus
   thwart error recovery when a button is accidentally bumped into
   such a location.
*/

int PlateZ( fibDB *fib, point *targetCoord, int *step, int boundryFlag )

{
   double softness;
   double targetRadius;

   /* in case this is invalid */

   *step = VOIDORD;


   /* set up for hard vs. soft limits */

   if( boundryFlag == HARD_BOUNDRIES )
      softness = 0.0;
   else
      softness = PLATE_BOUNDRY_SOFTNESS;

   /* compute the distance from the center of the plate */

   targetRadius = sqrt( ( targetCoord->x * targetCoord->x ) +
                        ( targetCoord->y * targetCoord->y ) );



   /* first see if we're withing the park/stow zone */

   if( ( targetRadius >= (STOW - 100) ) &&
       ( targetRadius <= (PARK + 1100) ) ){
      *step = PARKORD;
      return( PLATE_OK );	/* can't assign here... */
   }

   if( targetRadius <= ( PLATE  + softness ) ){
      *step = PLATEORD;
      return( PLATE_OK );
   }

   *step = VOIDORD;
   return( PLATE_BAD );
}


/*
		couldInteract

   Determine if the target buttons are extended far enough that
   they could interact.  This routine determines if *both* buttons
   are in a mutually accessible zone.  This is the primary (and
   fastest) method for eliminating the need for computationally
   intensive proximity tests.  Many potential interactions can
   be ruled out with this routine.
*/

int couldInteract( fibDB *fiber1, fibDB *fiber2 )

{
   double criticalRadius;

   int separationOrdinal;

   separationOrdinal = abs( fiber1->fiber_id - fiber2->fiber_id );

   if( separationOrdinal > ( MAXFIBER / 2 ) )
      separationOrdinal = abs( separationOrdinal - MAXFIBER );

   /* be sure target isn't parked or
      in a zone where no interaction could
      take place. */

#ifdef IGNORE
   /* this is the old version for the non-wiyn version */
   if( (fiber2->locked || ( fiber2->parkedStatus != EXTENDED ) ) &&
       ( separationOrdinal > PARKEDATRISK ) )
#endif

   if( fiber2->locked ){
      return(FALSE); 
   }

   criticalRadius = radiusList[ separationOrdinal ];

   if(  ( fiber1->CurGeometry->d4 > criticalRadius ) &&
        ( fiber2->CurGeometry->d4 > criticalRadius ) )

      return( TRUE );
   
   else
      return( FALSE );

}


/*
		buttonsConverge

   Determine if the buttons specified in the arguments
   converge towards each other, or diverge.  Retrun
   one of the following states:  Buttons diverge; 
   Buttons converge and are curved in opposite
   directions; buttons converge and are curved in the same
   direction.  Determine the low and hi fibers, with
   low  being the most CCW fiber wrt the angle between
   them.
*/

int buttonsConverge( fibDB *fib1, fibDB *fib2,
                     fibDB **lowFib, fibDB **hiFib,
		     double *thetaSep)

{
   int status;

   /* Figure out which fiber is the CW fiber on
      inside angle between the fibers. */

   if( abs( fib1->fiber_id - fib2->fiber_id ) > ( MAXFIBER / 2 )){

      /* the angle crosses the "zero" position
	 so the higher ordinal fiber becomes the "low"
	 one for the test of convergence */

      if( fib1->fiber_id > fib2->fiber_id ){
	 *lowFib = fib1;
	 *hiFib  = fib2;
      }
      else{
	 *lowFib = fib2;
	 *hiFib  = fib1;
      }
   }
   else{

      /* the angle doesn't cross the zero point, so
	 the lower ordinal is the lower fiber */

      if( fib1->fiber_id < fib2->fiber_id ){
	 *lowFib = fib1;
	 *hiFib  = fib2;
      }
      else{
	 *lowFib = fib2;
	 *hiFib  = fib1;
      }
   }
  
   /*  Now we can calculate the separation angle properly */

   *thetaSep = (*hiFib)->CurGeometry->theta - (*lowFib)->CurGeometry->theta ;

   /* normalize to +/- pi */

   while( *thetaSep < -M_PI )
      *thetaSep += TWOPI;

   while( *thetaSep > M_PI )
      *thetaSep -= TWOPI;

   *thetaSep = abs( *thetaSep );

   /* now check the signs of the fibers */

   if( *thetaSep <= M_PI_2 ){  /* test for an acute angle */

      /* are the fibers deflected away from each other? */

      if(( (*lowFib)->CurGeometry->phi < 0) &&
     	 ( (*hiFib)->CurGeometry->phi > 0 ) )
	 status = DIVERGE;

      else{  /*  fibers have same curvature or deflect towards each other */

	 /* if signs of deflection are opposite, they converge */

	 if(( (*lowFib)->CurGeometry->phi * (*hiFib)->CurGeometry->phi) <= 0)
	    status = CONVERGE;

	 else{  /* signs are the same so either one tells direction */

	    if( (*lowFib)->CurGeometry->phi < 0 )
	       status = PARALLELCW;
	    else
	       status = PARALLELCCW;
	 }
      }
   }

   else{  /* angle is obtuse */


      if( ( (*lowFib)->CurGeometry->phi > 0 ) &&
     	  ( (*hiFib)->CurGeometry->phi < 0 ) )
	 status = CONVERGE;
      else
	 status = DIVERGE;
   }

   return( status );
}



/*
		buttonHitsTube

   Determine if either button specified in the
   arguments hits either tube.  If the angle between
   the tubes is less than a pre-calculated value,
   there is only one possible interaction, and that is
   for the button farthest from the origin to hit
   the tube of the button closest to the origin.
*/


int buttonHitsTube( fibDB *targetFib,
                    fibDB *lowFib,
		    fibDB *hiFib,
		    int convergeStatus,
		    double sepAngle)

{

   struct flex_geometry *lowGeo = lowFib->CurGeometry;
   struct flex_geometry  *hiGeo =  hiFib->CurGeometry;

   double separation1;   /* Distance between tube of CCW button 
			    and CW tube */
   double separation2;   /* Distance between tube of CW button 
			    and CWW tube */

   double angle1 = sepAngle + hiGeo->phi;

   double angle2 = sepAngle - lowGeo->phi;

   int ouch = FALSE;

   separation1 = fiberButtonDistance( hiGeo, lowGeo, angle1 );

   separation2 = fiberButtonDistance( lowGeo, hiGeo, angle2 );

   /* check if both tubes clear buttons */

   if( (separation1 < BUTTONSPACE ) ){

      /* fputs( "separation 1 violation\n", stderr ); */
      
      if( hiFib ==targetFib ){
         ouch = TRUE;
      }
      else{
	 if( hiFib->parkedStatus == PARKED ){
	    /* create occultation */
	    occultAddPending( hiFib, lowFib );
	 }
	 else{
	    ouch = TRUE;
	 }
      }


   }

   if( (separation2 < BUTTONSPACE ) ){

      /* fputs( "separation 2 violation\n", stderr ); */
      
      if( lowFib ==targetFib ){
         ouch = TRUE;
      }
      else{
	 if( lowFib->parkedStatus == PARKED ){
	    /* create occultation */
	    occultAddPending( lowFib, hiFib );
	 }
	 else{
	    ouch = TRUE;
	 }
      }

   }



   return( ouch );

}


/*
		fiberButtonDistance

   Calculate the distance from a button to a tube.
   The first arg is the geometry for the button.
   The second arg is the geometry for the fiber.
   Return the distance between them as a point and
   a line, not accounting for the button or tube's
   size
*/


double fiberButtonDistance( struct flex_geometry *buttonGeometry, 
                            struct flex_geometry *fiberGeometry,
			    double angle )

{
   struct aLine *lineN;

   double distance;

   double dN = ( buttonGeometry->d2 * cos(angle)) - fiberGeometry->d3;

   /* above value is the point on the tube where we will
      calculate the slope and x,y coordinates to generate
      a line for distance testing */


   /* See if we are to far back on tube to use the 
      precalculated line through the button. */

   if( dN < 0 )	/* Buttons beyond end of tube */
     return( 2 * BUTTONSPACE );  /* guaranteed to make test pass */

   if( dN > MAXFORPRECALC ){

      lineN = allocLine();
      constructLine( fiberGeometry, dN, lineN );
      distance = pointLineSeparation( &buttonGeometry->button, lineN );
      freeLine(lineN);
   }

   else  /* Use the pre calculate line */
      distance = pointLineSeparation( &buttonGeometry->button,
     			 fiberGeometry->onePtAppx );


   return( distance );
}



/*
		constructLine

   Generate the line that is tangent to the tube at a 
   distance dN back from the button.
*/

void constructLine( struct flex_geometry *geometry, double dN, line *theLine )

{
   struct aPoint pointN;	/* working storage for point on tube */
   double localSlope;		/* slope of above point in local coords */
   double cartesianSlope;	/* above slope in cartesian coords */

   tubeSlope( geometry, dN, &localSlope, &cartesianSlope);
   tubeLocation( geometry, dN, &pointN);
   solveLinePointSlope( &pointN, cartesianSlope, theLine );
}

/*
		tubesCross

   Determine if the tubes specified in the arguments
   cross each other.
*/


int tubesCross( fibDB *targetFib,
                fibDB *lowFib,
		fibDB *hiFib,
		int   convergeStatus,
		double sepAngle)

{
   int ouch = 0;
   int i;
   int j;

   /*** KLUDGE ALERT****/

   /* there is a bug where opposed fibers occasionally report
      that they cross.  This is impossible since they can't
      go past the mid-point of the plate.  This fixes 
      the bug without figuring out what it is */

   if( abs( hiFib->fiber_id - lowFib->fiber_id ) == 144 )
      return( FALSE );

   /*** END KLUDGE ALERT****/

   for( i = 0; i < NUM_TUBE_SEGS; i++ ){
      for( j = 0; j < NUM_TUBE_SEGS; j++ ){
	 if( linesIntersect(lowFib->CurGeometry->segments[i],
			    hiFib->CurGeometry->segments[j] ) ){
	    ouch = TRUE;
	 }
      }
   }

   /* if there is an interaction, it doesn't count if one of the fibers is
      parked.  But it does mean that an occultation must be generated */

   if( ouch ){

      /* note that if "ouch" is true, they can't both be parked!
	 And if one or the other *isn't* parked, then we have a
	 true collision.  */

      if( ( hiFib == targetFib ) && ( lowFib->parkedStatus == PARKED ) ){
	 /* create occultation */
	 occultAddPending( lowFib, hiFib );
	 ouch = FALSE;
      }

      if( ( lowFib == targetFib ) && ( hiFib->parkedStatus == PARKED ) ){
	 /* create occultation */
	 occultAddPending( hiFib, lowFib );
	 ouch = FALSE;
      }
   }


   return( ouch );
}



/*
		miscOtherMagic

   Run a series of tests that are easy to see if we can
   rule out fiber interaction.  These include checking
   to see if a tube crosses the radial line of the
   target tube; and adding up the "shadow" angles of
   the buttons and comparing them to the separation
   angle.

   Note:  This routine depends on lowFib being the CW
   fiber on the angle between the emergence points of
   the two fibers.
*/

int miscOtherMagic( fibDB *lowFib, fibDB *hiFib, int convergeStatus, double sepAngle)

{

   int mysticalStateOfBeing = IGNORANCE; 

   switch( convergeStatus ){

      case CONVERGE    :

	       /*  hiFib has negative phi if we get here */
	       if( sepAngle > ( lowFib->CurGeometry->phi +
	                        lowFib->CurGeometry->shadow -
				hiFib->CurGeometry->phi+
				hiFib->CurGeometry->shadow) ){
		  
		  mysticalStateOfBeing = ENLIGHTENMENT; 
	       }
	       break;

      case PARALLELCCW :

	       if( sepAngle > ( lowFib->CurGeometry->phi +
	                        lowFib->CurGeometry->shadow) ){

		  mysticalStateOfBeing = ENLIGHTENMENT; 
	       }
      	       break;

      case PARALLELCW  :

	       if( sepAngle > ( -hiFib->CurGeometry->phi +
	                        hiFib->CurGeometry->shadow) ){

		  mysticalStateOfBeing = ENLIGHTENMENT; 
	       }
      	       break;

      default 	       :
      	       break;
   }

   return( mysticalStateOfBeing );
}



/*
		initRadiusList

   Generate the look-up table for the function couldInteract.
   This is basically a triangle solution with a correction
   added for the fact that the buttons are not lines and
   have width.
*/

void initRadiusList(void)

{
   double phi;
   double sinPhi;

   double radius;
   double correction;

   int i;

   for( i=1; i < (MAXFIBER / 2); i++ ){
      
      /* determine the angle between the emergence points
      of the fibers */

      /* This should *not* be 2 * M_PI in the numerator!
         This is deliberate!  We are using the half-angle. 
      */

      phi = i * M_PI / MAXFIBER; sinPhi = sin( phi );

      /* calculate the entry in the table */

      radius = ( sinPhi * PIVOT / sin( M_PI - phi - MAXANGLE ) );

      correction = ( BUTTONDIA / ( 2 * sin( phi + MAXANGLE )) );

      radiusList[i] = radius - correction;
   }

   /* set the last several entries to be the same.  Roundoff error
      seems to be corrupting higher order ones.  This is a reasonable
      trade-off
   */


   radiusList[ ( MAXFIBER / 2 - 1 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 2 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 3 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 4 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 5 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 6 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];
   radiusList[ ( MAXFIBER / 2 - 7 ) ] = radiusList[ ( MAXFIBER / 2 - 8 ) ];

}



/*
                buttonSeparation

   calculate the distance between the centers of the buttons
   on the fibers specified in the arguments then determine
   if there is sufficient space between the two buttons.
   Return zero if the buttons won't collide.
*/

int buttonSeparation( fibDB *fiber1, fibDB *fiber2 )

{
   double Dsep;

   double rb1 = fiber1->ButtonDia / 2;	/* radius of button 1 */
   double rb2 = fiber2->ButtonDia / 2;	/* radius of button 1 */


   Dsep = distance( &fiber1->CurGeometry->button, &fiber2->CurGeometry->button );

   if( ( Dsep - BUTTON_BUTTON_MARGIN ) < ( rb1 + rb2 ) ){
      return( TRUE );   /* they will collide */
   }
   else
      return( FALSE );	/* they won't collide*/
}



/*
                overButton

   See if the specified point is over the specified button.

*/

int overButton( fibDB *fiber1, point *thePoint )

{
   double Dsep;

   double rb1 = fiber1->ButtonDia / 2;	/* radius of button 1 */

   Dsep = distance( &fiber1->CurGeometry->button, thePoint );

   if( Dsep < rb1 ){
      return( TRUE );   /* they will collide */
   }
   else
      return( FALSE );	/* they won't collide*/
}



/*
		tubeSlope

   find the slope of a point on the specified fiber, "fib",
   with the specified distance, "dN", from the button along the
   radial line to the point of emergence.
*/

void tubeSlope( struct flex_geometry *flex,
                double dN, 		/*  radial distance from button to point N on tube */
		double *localSlope, 	/* slope in the local coordinate system */
		double *cartesianSlope) /* slope in the primary coordinate system */

{
      *localSlope = flex->c / 2 * ( (flex->d1 * flex->d1 ) - ( dN * dN ) );

      *cartesianSlope = tan( atan( *localSlope ) + flex->theta );
}



/*
		tubeLocation

   find the coordinates of a point on the specified fiber, "fib",r
   with the specified distance, "dN", from the button along the
   radial line to the point of emergence.
*/

void tubeLocation( struct flex_geometry *flex, double dN, point *thePoint )

{
   double rN;
   double deflN;	/*  deflection from radial line and point N on tube*/
   double phiN;

   deflN = flex->c * (  ( dN * dN * dN ) / 6.0  -
			(flex->d1 * flex->d1 ) * dN / 2.0 +
			(flex->d1 * flex->d1 * flex->d1) / 3.0 );

   rN = sqrt( (flex->d3 + dN) * (flex->d3 + dN) +
	      ( deflN * deflN ) );

   if( deflN == 0 )
      phiN = 0;
   else
      phiN = atan2( deflN, (flex->d3 + dN) );

   thePoint->x =  rN * cos( flex->theta - phiN );

   thePoint->y =  rN * sin( flex->theta - phiN );
}




/*
		tubeAngle

   Find the angle between the radial line and the tube
   at a the specified point on the tube.
*/

void tubeAngle( struct flex_geometry *flex, double dN, double *phiN )

{
   double deflN;	/*  deflection from radial line and point N on tube*/

   deflN = flex->c * (  ( dN * dN * dN ) / 6.0  -
		     (flex->d1 * flex->d1 ) * dN / 2.0 +
		     (flex->d1 * flex->d1 * flex->d1) / 3.0 );

   *phiN = -atan2( deflN, (flex->d3 + dN) );
}



/*
		magicNumbers

   Take the fiber who's data base is pointed to by fib and calculate
   the magic numbers for tube flexure and tube position geometry for the
   target location x,y.

   Put the results into the structure pointed to by results.
   This routine will be called before a tube is moved as part of
   verifying that a move is valid.  If the move is valid, it will occur,
   and the structure pointed to by results will be linked to the
   fiber's database and the old flex_geometry structure freed.
   If the move is not valid,  the current flex_geometry structure
   is left intact.

   These variable names are rather cryptic.  Could be improved, but not
   by much.  This really requires a separate document with graphics to
   explain what is going on.

*/


void MagicNumbers( struct flex_geometry *results )

{
   double tmpAngle;
   int i;
   struct aPoint tempPoint;

   /* calculate the angle of the radial line wrt 0 */

   results->theta = atan2( results->pivot.y, results->pivot.x );

   /* calculate the distance from the button to the origin */

   results->d2    = sqrt(  results->button.x * results->button.x +
  			 results->button.y * results->button.y );

   /* calculate the angle at the origin between the button
      and the radial line, retaining sign convention even
      when the angle crosses the +/- PI boundry. */

   if( (results->button.x == 0) && (results->button.y == 0) ){

      results->phi = 0;
   }

   else{

      tmpAngle = atan2( results->button.y, results->button.x );

      if( results->theta >= 0 ){
	 if( tmpAngle <= 0 )
	    tmpAngle += TWOPI;
      }
      else
	 if( tmpAngle  >= 0 )
	    tmpAngle -= TWOPI;

      results->phi = tmpAngle - results->theta;

      if( results->phi < -M_PI )
	 results->phi += TWOPI;

      if( results->phi > M_PI )
	 results->phi -= TWOPI;
   }

   results->d3    = results->d2 * cos( results->phi );

   results->d1    = PIVOT - results->d3 ;

   results->defl  = results->d2 * sin( results->phi );

   results->psi   = atan2( results->defl, results->d1 );

   results->d4    = sqrt( results->defl * results->defl +
  			  results->d1 * results->d1 );

   results->c     =  -( results->defl /
  		 ( results->d1 * results->d1 * results->d1 / 3.0 ) );

   results->buttonLocalSlope =  results->d1 * results->d1 * results->c / 2;

   results->chi   = atan( results->buttonLocalSlope );

   /* calculate the angle at origin shadowed by the button 
      (center to edge of button.  angle of whole button'
      is double this value) */

   if( results->d2 < ( BUTTONDIA / 2 ) )
      results->shadow = M_PI;
   else
      results->shadow =  2 * asin( ( (double)BUTTONDIA / 2 ) / results->d2 );

   /* solve a line that runs through the button and has the same
   slope as the tube where it emerges from the button. */

   tmpAngle = results->theta + results->chi;

   if( tmpAngle != M_PI_2 )
      results->buttonCartSlope =  tan( tmpAngle );
   else
      results->buttonCartSlope =  HUGE_VAL;	/* infinity */
   
   solveLinePointSlope( &results->button, results->buttonCartSlope,
  			 results->onePtAppx );

   /* solve a line that runs from the button to the pivot point */

   solveLineCartesian( &results->button, &results->pivot, results->twoPtAppx );

   /*** represent the tube with four straight line segments ***/

   /* set the end points that need not be calculated */

   *results->segments[0]->p1 = results->button;	/* seg. end that touches button.*/

   *results->segments[(NUM_TUBE_SEGS - 1)]->p2 = results->pivot;

   /* Now find the points for the rest of the segment ends */

   for( i = 0; i < ( NUM_TUBE_SEGS - 1); i++ ){

      /* Find location of the tube at the pair of segment ends. */

      tubeLocation( results, ( results->d1 * segConstants[ i + 1 ] ), &tempPoint );

      *results->segments[i]->p2 = tempPoint;
      *results->segments[i +1]->p1 = tempPoint;
   }

   for( i = 0; i < NUM_TUBE_SEGS; i++ ){

      solveLineCartesian(
         results->segments[i]->p1,
         results->segments[i]->p2,
	 results->segments[i] );
   }


}

void initDbase(void)

{

   struct fiber_dbase *pFib;
   int i;

   double angle = 0;
   double dangle = ( M_PI * 2 ) /  MAXFIBER ;

   /* This next value compensates for the fact that the stow
      positions are not lined up with the pivots and park positions. */

   double tweak	= TWOPI / 400000.0;

#ifdef VERBOSE
      fprintf( stderr, "initializing:     " );
#endif

   for(i=0;i<MAXFIBER;i++){

#ifdef VERBOSE
      fprintf( stderr, "\b\b\b%3d", i );
#endif

      pFib = &FiBase[i];

      pFib->occultedBy   = occultAlloc( (struct fiber_dbase *)NULL );
      pFib->weOccult     = occultAlloc( (struct fiber_dbase *)NULL );
      pFib->pendOccultBy = occultAlloc( (struct fiber_dbase *)NULL );
      pFib->pendWeOccult = occultAlloc( (struct fiber_dbase *)NULL );

      angle =  i * dangle;

      pFib->fiber_id     = i;

      pFib->PivotPos.x   = PIVOT * cos( angle + tweak );
      pFib->PivotPos.y   = PIVOT * sin( angle + tweak );

      pFib->ParkPos.x    = PARK * cos( angle + tweak );
      pFib->ParkPos.y    = PARK * sin( angle + tweak );

      pFib->StowPos.x    = STOW * cos( angle + tweak + (STOWOFFSET) );
      pFib->StowPos.y    = STOW * sin( angle + tweak + (STOWOFFSET) );

      pFib->MaxExtend    = MAXEXTEND;
      pFib->MaxAngle     = MAXANGLE;

      pFib->ButtonHeight = BUTTONHEIGHT;
      pFib->GrabHeight   = GRABHEIGHT;
      pFib->ButtonDia    = BUTTONDIA;

      pFib->TubeDia      = TUBEDIA;

      pFib->CurPos.x     = pFib->StowPos.x;
      pFib->CurPos.y     = pFib->StowPos.y;


      /* allocate and initialize the flexure geometry structure */

      if( ( pFib->TrialGeometry = allocGeometry() ) == NULL ){
	 fputs( "ERROR:  Malloc failed for geometry structure\n", stderr );
	 exit( 1 );
      }
	 
      if( ( pFib->CurGeometry = allocGeometry() ) == NULL ){
	 fputs( "ERROR:  Malloc failed for geometry structure\n", stderr );
	 exit( 1 );
      }

      /* doing it twice insures that both geometrys get set! */

      setLocation( pFib, &pFib->StowPos, 0 /* don't update file! */ );
      setLocation( pFib, &pFib->StowPos, 0 /* don't update file! */ );

      pFib->parkedStatus  = STOWED;
      pFib->futureParkedStatus  = PARKED;

      pFib->outOfDate  = FALSE;
      pFib->tweaked    = FALSE;

   }

#ifdef VERBOSE
   fputs( "\n", stderr );
#endif

}


/*
		setDbase

   Set all the fibers to either the stowed or parked positions.

   The reason that the calls to setLocation are made twice is to
   insure that the both geometry structures are identical and
   set to a known location.
*/

void setDbase( int flag )

{
   struct fiber_dbase *pFib;
   int i;

   for(i=0;i<MAXFIBER;i++){

      pFib = &FiBase[i];

      if( flag ){
	 setLocation( pFib, &pFib->StowPos, UPDATE_FILE );
	 pFib->parkedStatus  = STOWED;
      }

      else{
	 setLocation( pFib, &pFib->ParkPos, UPDATE_FILE );
	 pFib->parkedStatus  = PARKED;
      }
   }
}



/*
		setLocation
   
   Update the database for the given fiber to show
   that it is at the specified location without actually
   moving it.  Used for error recovery.
*/

void setLocation( fibDB *fib, struct plate_coords *location, int archiveFlag )


{
   struct aPoint target;

   /* convert to double */

   target.x = location->x;
   target.y = location->y;

   prepNewLocation(fib, &target );
   recordNewLocation(fib, archiveFlag );
}





/*		restoreOldLocation

   Clear the fiber inTransit.
 
*/

void restoreOldLocation()

{
   if( ( FiberInTransit.fiber != NULL) &&
       ( FiberInTransit.state != IN_TRANSIT) ){
      FiberInTransit.fiber = (struct fiber_dbase *)NULL;
      FiberInTransit.state = STATIC;
   }
}



/*
		prepNewLocation

   Prepare the TrialGeometry of the specified fiber for the
   specified location.  If this geometry is actually used,
   it will be linked into the CurGeometry for the fiber and
   the old CurGeometry "recycled" and linked in as the Trial
   Geometry.  recordNewLocation performs the swap of geometry
   structures.
*/

int prepNewLocation( fibDB *fib, point *target )

{
   struct flex_geometry *geo = fib->TrialGeometry;

 /*
   FiberInTransit.fiber = fib;
 */

   geo->button.x = target->x;
   geo->button.y = target->y;

   /* this could be moved to the initialization */

   geo->pivot.x = fib->PivotPos.x;
   geo->pivot.y = fib->PivotPos.y;

   MagicNumbers( geo );

   /* don't change the state unless the transition would
      be from static to move pending.  This routine may
      be called when a move fails mid-stream, and in such
      a case, we want to retain the current state 
   */

   if( FiberInTransit.state == STATIC )
      FiberInTransit.state = MOVE_PENDING;

   return( FALSE );	/* no error */
}



/*
		recordNewLocation

   Record a successful move by updating the location_now field
   and recording the new location in the disk archive.  Finally,
   call clearOldLocation to dispose of the old geometry information.
*/

void recordNewLocation( fibDB *fib, int archiveFlag )

{
   struct flex_geometry *tempGeometry;

   /* Update the current position with the coordinates that
      were used for target and collision testing. */

   fib->CurPos.x = fib->TrialGeometry->button.x;
   fib->CurPos.y = fib->TrialGeometry->button.y;

   /* save a link to the memory used for the old info */

   tempGeometry = fib->CurGeometry;
   
   /* link in the new info */

   fib->CurGeometry = fib->TrialGeometry;

   /* link the original memory back in for future use */

   fib->TrialGeometry = tempGeometry;

   /* park and stow will override the line below */

   fib->parkedStatus = EXTENDED;


   if( archiveFlag ){

#ifdef ARCHIVE
   archivePosition( fib, positionArchive );
   recordNewOccultations( fib );
#endif
   }

}



/*
                allocGeometry

   allocate a "flex_geometry" structure, and return a pointer to it.
*/

struct flex_geometry *allocGeometry( void )

{
   int i;

   struct flex_geometry *tempGeo;

   tempGeo = (struct flex_geometry *)malloc(
       (unsigned)sizeof( struct flex_geometry ));

   if( tempGeo != NULL ){

      tempGeo->twoPtAppx = allocLine();
      tempGeo->onePtAppx = allocLine();

      for( i = 0; i < NUM_TUBE_SEGS; i++ ){

	 tempGeo->segments[i] = allocLine();

	 tempGeo->segments[i]->p1 = allocPoint();
	 tempGeo->segments[i]->p2 = allocPoint();
      }

   }

   tempGeo->button.x = 0;
   tempGeo->button.y = 0;

   return( tempGeo );
}


/*
		freeGeometry

   free a "flex_geometry" structure and it's associated aLine structures.
*/

void freeGeometry(struct flex_geometry *pGeo)

{
   int i;

   if( pGeo != NULL ){

      if( pGeo->twoPtAppx != NULL )
	 freeLine( pGeo->twoPtAppx );

      if( pGeo->onePtAppx != NULL )
	 freeLine( pGeo->onePtAppx );

      for( i = 0; i < NUM_TUBE_SEGS; i++ )
	 if( pGeo->segments[i] != NULL )
	    freeLine( pGeo->segments[i] );

      free( (char *) pGeo );
   }
}


